#!/usr/bin/perl
# A Perl text editor.
# Copyright (c) 1998 by unSlider, all rights reserved.
# You may use and distribute this under the terms of the Perl Artistic License.

use Curses;

$acedver=19981223.0;		# our version number
@buffer=();
$cbuf=\@buffer;			# current buffer
%info=();
$cinfo=\%info;			# settings that apply to current buffer

%defbindings=(			# default key bindings
	      "\cW" => \&quit,
	      "\c[" => sub {
		  &dometa($$cinfo{metabindings});
	      },
	      "\cf" => \&cursfwd,
	      @{[KEY_RIGHT]} => \&cursfwd,
	      "\cb" => \&cursback,
	      @{[KEY_LEFT]} => \&cursback,
	      "\ca" => \&cursbol,
	      @{[KEY_HOME]} => \&cursbol,
	      "\ce" => \&curseol,
	      @{[KEY_END]} => \&curseol,
	      "\cm" => \&insnewline,
	      "\ch" => \&backspace,
	      @{[KEY_BACKSPACE]} => \&backspace,
	      "\cd" => \&delunder,
	      @{[KEY_DC]} => \&delunder,
	      "\cn" => \&cursdown,
	      @{[KEY_DOWN]} => \&cursdown,
	      "\cp" => \&cursup,
	      @{[KEY_UP]} => \&cursup,
	      "\cv" => \&curspgdn,
	      @{[KEY_NPAGE]} => \&curspgdn,
	      "\cy" => \&curspgup,
	      @{[KEY_PPAGE]} => \&curspgup,
	      "\ck" => \&cutline,
	      "\cu" => sub {&paste($cutbuffer)},
# now we can handle embedded tabs
#	      "\ci" => \&tab,
	   );

%defmetabindings=(		# default meta-key bindings
		  "name" => "meta",
		  "z" => sub {&statmsg("meta-z was pressed");},
		  "c" => sub {
		      eval(&getline("> "));
		      if($@) {
			  &errmsg($@);
		      }
		      &movecurs();
		  },
		  "f" => \&find,
		  "s" => \&saveas,
		  "l" => \&load,
		  "w" => sub {
		      if($$cinfo{wordwrap}) {
			  $$cinfo{wordwrap}=0;
			  &statmsg("Disabled word-wrap");
		      } else {
			  $$cinfo{wordwrap}=1;
			  &statmsg("Enabled word-wrap");
		      }
		  },
		  "\cA" => \&cursbof,
		  "\cE" => \&curseof,
	       );

$info{bindings}=\%defbindings;
$info{metabindings}=\%defmetabindings;
$info{wintop}=0;
$info{statlines}=1;
$info{statfunc}=\&statbar;
# cursy is the offset into the CURRENTLY DISPLAYED WINDOW of the current line
# position is the offset of the first line of the window into the buffer
# cursx is the offset, IN FILE (not display) CHARACTERS, into the BUFFER VIEW
# OF THE CURRENT LINE (ie, left/right panning does not figure in)
# pan is the number of characters skipped at the beginning of each line
$info{cursx}=0;			# x-position of the cursor (insertion point)
$info{cursy}=0;			# y-position of the cursor
$info{position}=0;		# line num in buffer on the top of the screen
$info{pan}=0;			# offset from the left of the screen
$info{tabstop}=5;		# number of spaces in a tab stop

initscr;
cbreak;				# disable buffering and erase/kill
# once we're stable enough to ignore ^C we should probably go raw
nonl;				# no newline->\n munging
noecho;				# disable getch echoing
timeout(-1);			# getch blocks for input
keypad(stdscr,1);		# decode extended keys to curses constants

# $LINES doesn't get a value until after we initscr
$info{winlines}=$LINES-2;

&{$$cinfo{statfunc}};
$diedie=0;

if(-e '/etc/acedrc') {
    &source('/etc/acedrc');
}
if(-e "$ENV{HOME}/.acedrc") {
    &source("$ENV{HOME}/.acedrc");
}

$startline=0;
while(defined($_=shift)) {
    if($_=~/^\+(\d+)/) {
	$startline=$1;
    } else {
	$startopen=$_;
    }
}
if(defined($startopen)) {
    &load($startopen);
    &zaptoline($startline);
}

while($diedie==0) {
    $inchar=getch;
    if($inchar eq $lastchar) {
	$repeated=1;
    } else {
	$repeated=0;
    }
    if(defined($$cinfo{bindings}{$inchar})) { # key is bound to a function
	&{$$cinfo{bindings}{$inchar}};
    } else {			# no binding, just insert this character
	&insatcurs($inchar);
	refresh;
    }
    $lastchar=$inchar;
}
endwin;

# draw the status bar
sub statbar {
    addstr($$cinfo{winlines}+$$cinfo{wintop},0,
	   "--aced [$$cinfo{filename}]--");
    clrtoeol;
    &movecurs();
}

# draw a message under the status line
sub statmsg {
    addstr($LINES-1,0,$_[0]);
    clrtobot;
    &movecurs();
}

# display an error message - statmsg but with a beep
sub errmsg {
    beep;
    &statmsg($_[0]);
}

# get a line of user input outside the edit region
# optional argument: prompt
sub getline {
    my $line;
    my $y;

    echo;
    move($LINES-1,0);
    if(defined($_[0])) {
	addstr($_[0]);
    }
    clrtobot;
    getstr($line);
    noecho;
    
    return $line;
}

# determine what line of the buffer we're on
sub bufline {
    return $$cinfo{cursy}+$$cinfo{position};
}

# insert character(s) at the cursor and advance the cursor
sub insatcurs {
    my($bar,$foo);

    substr($$cbuf[&bufline],$$cinfo{cursx})=
	$_[0].substr($$cbuf[&bufline],$$cinfo{cursx});
    $$cinfo{cursx}+=length($_[0]);
    if($$cinfo{wordwrap} && length($$cbuf[&bufline])>$COLS-5) {
	&wrapwords();
    } else {
#	&refreshbuf(); # fixme only needs to refresh this line
	&refreshline($$cinfo{cursy});
    }
}

# refresh the image of the buffer on the screen
sub refreshbuf {
    my $curstr;
    my $i;

#    $dontrefresh && return;

    for($i=0;$i<$$cinfo{winlines};$i++) {
# this substr() here is ugly but if we defer to the addstr() line it becomes
# a lhs context which means perl will complain if requested offset is beyond
# the end of the string
	$curstr=substr(
		       &escchars($$cbuf[$i+$$cinfo{position}]),
		       $$cinfo{pan},$COLS);
	addstr($i+$$cinfo{wintop},0,$curstr);
	clrtoeol;
#	&statmsg(length($curstr)." ".defined($curstr)." ".$$cinfo{pan});
    }
# note that sometimes &movecurs can perhaps reenter refreshbuf? this might
# happen if we manage to get here with curx>$COLS. but this is a good thing.
# movecurs will call refreshbuf, but the next movecurs call won't, so we only
# recurse one deep. this has the benefit that refreshing the screen causes it
# to pan as appropriate.
# actually.. there is a case where it'll recurse more than one deep. if you
# have a line over $COLS*2 characters long, and you're at the end, and you
# scroll to the left. it'll reset pan to 0 and then increment it by 5 each
# call until the cursor is visible. this is ok but it's bad for performance.
# to fix this we should make movecurs smarter about moving backwards.
    &movecurs();
}

# Refresh one line on the screen
sub refreshline {
    my $curstr;

    $curstr=substr(
		   &escchars($$cbuf[$_[0]+$$cinfo{position}]),
		   $$cinfo{pan},$COLS);
    addstr($_[0]+$$cinfo{wintop},0,$curstr);
    clrtoeol;
    &movecurs();
}

# move the cursor 1 character forward
sub cursfwd {
    if($$cinfo{cursx}==length($$cbuf[&bufline])) {
	if(defined($$cbuf[&bufline+1])) {
	    $$cinfo{cursx}=0;
	    &cursdown();
	} else {
	    &errmsg("End of buffer");
	}
    } else {
	$$cinfo{cursx}++;
	&movecurs();
    }
}

# move the cursor 1 character back
sub cursback {
    if($$cinfo{cursx}==0) {
	if($$cinfo{cursy}>0) {
	    &cursup();
	    $$cinfo{cursx}=length($$cbuf[&bufline]);
	} else {
	    &errmsg("Beginning of buffer");
	}
    } else {
	$$cinfo{cursx}--;
    }
    &movecurs();
}

# move the cursor to the beginning of the current line
sub cursbol {
    $$cinfo{cursx}=0;
    &movecurs();
}

# move the cursor to the end of the current line
sub curseol {
    $$cinfo{cursx}=length($$cbuf[&bufline]);
    &movecurs();
}

# split the current line at the cursor
sub insnewline {
    splice(@$cbuf,&bufline+1,0,substr($$cbuf[&bufline],$$cinfo{cursx}));
    substr($$cbuf[&bufline],$$cinfo{cursx})='';
    &refreshbuf;
    &cursfwd;
}

# insert string(s) at the cursor, interpreting newlines properly
# multiple arguments are concatenated so you can do &paste(`ls`)
sub paste {
#    my $char;
    my @lines;
    my $line;

    @lines=split(/\n/,join('',@_),-1);

#    &statmsg("$#lines");
    while(defined($line=shift(@lines))) {
	&insatcurs($line);
#	&insatcurs("[$#lines]");
	if($#lines>=0) {
	    &insnewline();
	}
    }

    &refreshbuf();
}

# delete the character to the left of the cursor
sub backspace {
    if($$cinfo{cursx}==0) {
	if(&bufline==0) {
	    &errmsg("Beginning of buffer");
	} else {
	    $$cinfo{cursx}=length($$cbuf[&bufline-1]);
	    splice(@$cbuf,&bufline-1,2,$$cbuf[&bufline-1].$$cbuf[&bufline]);
	    $$cinfo{cursy}--;
	}
    } else {
	substr($$cbuf[&bufline],$$cinfo{cursx}-1)=
	    substr($$cbuf[&bufline],$$cinfo{cursx});
	$$cinfo{cursx}--;
    }
    &refreshbuf();
}

# delete the character under the cursor
sub delunder {
    if($$cinfo{cursx}==length($$cbuf[&bufline])) {
	if(defined($$cbuf[&bufline+1])) {
	    splice(@$cbuf,&bufline,2,$$cbuf[&bufline].$$cbuf[&bufline+1]);
	} else {
	    &errmsg("End of buffer");
	}
    } else {
	substr($$cbuf[&bufline],0,$$cinfo{cursx}+1)=
	    substr($$cbuf[&bufline],0,$$cinfo{cursx});
    }
    &refreshbuf();
}

# Move the cursor down one line
sub cursdown {
    if(defined($$cbuf[&bufline+1])) {
	&zaptoline(&bufline+1);
    } else {
	&errmsg("End of buffer");
    }
}

# Move the cursor up one line
sub cursup {
    if(&bufline==0) {
	&errmsg("Beginning of buffer");
    } else {
	&zaptoline(&bufline-1);
    }
}

# Save the current buffer. If an argument is given, use that as the filename,
# else prompt the user. If an argument is given, prompt before overwriting.
# FIXME we should avoid race conditions when writing to /tmp and such
sub saveas {
    my($filename, $sure, $line);
    my @foo;

    if(defined($_[0])) {
	$filename=$_[0];
    } else {
	$filename=&getline("Filename: ");
    }

    if(-e $filename) {
	$sure=&getline("Sure you want to overwrite $filename? ");
	if($sure!~/^\s*[Yy]/) {
	    &statmsg("Cancelled");
	    return;
	}
    }

    open(FOO,">$filename");
    $,="\n";
    print FOO @$cbuf;
    undef($,);
#    foreach $line (@$cbuf) {
#	print FOO "$line\n";
#    }
    close FOO;

    $$cinfo{filename}=$filename;

    @foo=@$cbuf;            # make a copy of cbuf so we can check for mods
    $$cinfo{origbuf}=\@foo;
}

# Execute a subcommand, either by suspending curses or in an xterm.
sub shell {
    my $command;

    if($_[0]) {
	$command=$_[0];
    } else {
	$command="$ENV{SHELL} -i";
    }

    if($ENV{DISPLAY}) {
	system("xterm -e $ENV{SHELL} -c '$command' &");
    } else {
	endwin;
	system($command);
	refresh;
    }
}

# Scroll the buffer up a page
sub curspgup {
    my $newpos;
    
    if($$cinfo{cursy}>0) {
	$$cinfo{cursy}=0;
	&movecurs();
    } else {
	&cursup();
#	$newpos=&bufline-$$cinfo{winlines};
#	if($newpos<0) {
#	    &zaptoline(0);
#	} else {
#	    &zaptoline($newpos);
#	    $$cinfo{cursy}=$$cinfo{winlines}-1;
#	    &movecurs();
#	}
    }
}

# Scroll the buffer down a page
sub curspgdn {
    if($$cinfo{cursy}<$$cinfo{winlines}-1) {
	$$cinfo{cursy}=$$cinfo{winlines}-1;
	if(&bufline>$#$cbuf) {
#	    $$cinfo{cursy}=0; # FIXME
	    $$cinfo{cursy}=0;
	    $$cinfo{cursy}+=$#$cbuf-&bufline;
	}
	&movecurs();
    } else {
	&cursdown();
    }
}

# Erase the current line and put it in the cut buffer
sub cutline {
    if(!$repeated) {
	$cutbuffer='';
    }

    if($$cbuf[&bufline]=~/^$/) {
	&delunder();
	$cutbuffer.="\n";
    } else {
	$cutbuffer.=$$cbuf[&bufline];
	$$cbuf[&bufline]='';
	$$cinfo{cursx}=0;
	&refreshbuf();
    }
}

# indent to a multiple of 5 characters
# (going away; replace with &insatcurs("\cI"))
sub tab {
baz:for(;;) {
	&paste(" ");
	($$cinfo{cursx}=~/[05]$/) and last baz;
    }
}

# load a file. if an argument is given, use it. else prompt.
sub load {
    my $filename;
    my $convertp;
    my $chomped;
    my @foo;

    if(defined($_[0])) {
        $filename=$_[0];
    } else {
        $filename=&getline("Filename: ");
    }

    if(!-e $filename) {
	&errmsg("No such file");
    } else {
	open(FOO,"<$filename");
	@$cbuf=<FOO>;
# if file ends in a \n, add a blank line to the end of the buffer so that
# the file's trailing \n is preserved on save
	if($$cbuf[$#cbuf]=~/\n$/) {
	    push @$cbuf,'';
	    # if this is an msdos style file, the chomp in the converter won't
	    # chomp anything from the last line, so we need to let it know that
	    # the chomp count should be expected to be 1 more than expected
	}
	chomp @$cbuf;
	if($$cbuf[0]=~/\r$/) {	# if it looks like \r\n, strip all last chars
	    $convertp=&getline("This file looks like MS-DOS \\r\\n style. Strip the \\r's? ");
	    if($convertp=~/^\s*[Yy]/) {
		&statmsg("Stripping trailing characters from MS-DOS style file.");
		$/="\r";
		$chomped=chomp(@$cbuf);
		$/="\n";
		# the last line of @$cbuf will never have a \r: either the file
		# didn't end in a newline, or it did and we added a blank (no
		# \r) line at the end. so we'll always chomp one less than the
		# 1-based number of lines in the file if it's truly a \r\n
		# style of file, which happs to be equal to the 0-based number
		# of lines in the file.
		if($chomped!=$#$cbuf) {
		    &statmsg(($#$cbuf-$chomped).
			     " lines didn't end in \\r\\n! Got rid of the other \\r's anyway.");
		}
	    } else {
		&statmsg("Not reformatting file.");
	    }
	}
	close FOO;

	@foo=@$cbuf;		# make a copy of cbuf so we can check for mods
	$$cinfo{origbuf}=\@foo;

	$$cinfo{cursx}=0;
	$$cinfo{cursy}=0;
	$$cinfo{position}=0;
	$$cinfo{pan}=0;
	$$cinfo{filename}=$filename;
	&refreshbuf();
	&{$$cinfo{statfunc}};
    }
}

# find a regex in the current buffer
sub find {
    my $regex;
    my $i;
    my $curend;

    if(defined($_[0])) {
        $regex=$_[0];
    } else {
        $regex=&getline("Regex: ");
    }

    # check current line after cursor
    if($$cinfo{cursx}<length($$cbuf[&bufline])) {
	$curend=substr($$cbuf[&bufline],$$cinfo{cursx}+1);
	if($curend=~/$regex/) {
	    &findfound(&bufline,index($$cbuf[&bufline],$&,$$cinfo{cursx}+1));
#	    &statmsg("Found on current line");
	    return;
	}
    }

    # check file after this line
    for($i=&bufline+1;defined($$cbuf[$i]);$i++) {
	if($$cbuf[$i]=~/$regex/) {
	    &findfound($i,index($$cbuf[$i],$&));
#	    &statmsg("Found after current line");
	    return;
	}
    }

    # check file before this line
    # also check this line (we ignored the beginning part before)
    for($i=0;$i<=&bufline;$i++) {
	if($$cbuf[$i]=~/$regex/) {
	    &statmsg("Search wrapped");
	    &findfound($i,index($$cbuf[$i],$&));
#	    &statmsg("Found before current line");
	    return;
	}
    }

    &statmsg("Not found")
}

# called when &find() finds a string: &findfound(line,position)
sub findfound {
#    &statmsg("found at $_[0] $_[1]");
    &zaptoline($_[0]);
    $$cinfo{cursx}=$_[1];
    &refreshbuf();
}

# move the cursor to the specified line, scrolling the buffer if needed
sub zaptoline {
    my $rb=0;			# whether to refreshbuf

    if($_[0]<$$cinfo{position}) { # before top of screen
	if($$cinfo{position}-$_[0]<=($$cinfo{winlines}/2)) {
				# half a screen or less above the top
	    $$cinfo{position}-=int($$cinfo{winlines}/2);
	} else {
				# more than half a screen above the top
	    $$cinfo{position}=$_[0]-2; # give some context
	}
	if($$cinfo{position}<0) {
	    $$cinfo{position}=0;
	}
	$rb=1;
    } elsif ($_[0]>=$$cinfo{position}+$$cinfo{winlines}) { # below bottom
	if($_[0]-($$cinfo{position}+$$cinfo{winlines}) < ($$cinfo{winlines}/2))
	   {			# half a screen or less below the bottom
	       $$cinfo{position}+=int($$cinfo{winlines}/2);
	       if($$cinfo{position}>$#$cbuf) {
		   $$cinfo{position}=$#$cbuf;
	       }
	} else {
				# more than half a screen below the bottom
	    $$cinfo{position}=$_[0]-2; # give some context
	    if($$cinfo{position}<0) {
		$$cinfo{position}=0;
	    }
	}
	$rb=1;
    }				# else on the current screen

    $$cinfo{cursy}=$_[0]-$$cinfo{position};

    if($$cinfo{cursx}>length($$cbuf[&bufline])) {
	$$cinfo{cursx}=length($$cbuf[&bufline]);
    }

    if($rb) {
	&refreshbuf();
    } else {
	&movecurs();
    }
}

# Move the cursor to the top of the file
sub cursbof {
    $$cinfo{cursx}=0;
    &zaptoline(0);
}

# Move the cursor to the eof
sub curseof {
    &zaptoline($#$cbuf);
    $$cinfo{cursx}=length($$cbuf[&bufline]);
}

# Given a string, escape non-printables
sub escchars {
    my $foo=$_[0];
    my $tabchar;

    $foo=~s/([\x00-\x08\x0A-\x1F])/&esc1char($1)/ge;
    while(($tabchar=index($foo,"\cI"))>=0) {
	substr($foo,$tabchar,1)=
	    " "x($$cinfo{tabstop}-($tabchar%$$cinfo{tabstop}));
    }
    return $foo;
}

# Convert one control character to its printable equivalent
sub esc1char {
    return '\x'.sprintf("%02X",ord($_[0]));
}

# Put the cursor in the right place
sub movecurs {
    my $x;

    $x=length(&escchars(substr($$cbuf[&bufline],0,$$cinfo{cursx})));
    if($x-$$cinfo{pan}<0) {
	$$cinfo{pan}=0;
	&refreshbuf();
    } elsif($x-$$cinfo{pan}>=$COLS) {
	$$cinfo{pan}+=5;
	&refreshbuf();
    }
    move($$cinfo{cursy}+$$cinfo{wintop},$x-$$cinfo{pan});
    refresh;
}

# word-wrap this line and any lines pushed over the edge by wrapping this one
sub wrapwords {
    my $i=&bufline;

    while(length($$cbuf[$i])>$COLS-5) {
	if(($bar,$foo)=$$cbuf[$i]=~/(.*\s)(.+)/) {
	    $$cbuf[$i]=$bar;
#	    if(defined($$cbuf[$i+1])) {
		$$cbuf[$i+1]=$foo.$$cbuf[$i+1];
#	    } else {
#		splice(@$cbuf,$i+1,0,$foo);
#	    }
	    if($$cinfo{cursx}>length($$cbuf[&bufline])) {
		$$cinfo{cursx}-=length($$cbuf[&bufline]);
		&cursdown();
	    }
	}
	$i++;
    }
    &refreshbuf();
}

# read one key and use a different set of bindings for it. argument is a ref
# to the binding hash for this meta-key
sub dometa {
    my $inchar;
    my $bindings=$_[0];

    halfdelay(20); # return ERR if no input for 2 sec
    $inchar=getch;
# unsetting halfdelay is a pain in the ass
    nocbreak;
    cbreak;

    if($inchar==ERR) {
	&statmsg("$$bindings{name}-");
	$inchar=getch;
    }

    if($repeated && ($inchar ne $$bindings{lastmeta})) {
	$repeated=0;
    }
    if(defined($$bindings{$inchar})) {
	&{$$bindings{$inchar}}();
    } else {
	&errmsg("no such $$bindings{name} binding $inchar");
    }
    $$bindings{lastmeta}=$inchar;
}

# Evaluate a perl source file.
sub source {
    do $_[0]; #FIXME some errors/warnings can blurt straight to the screen
    if($@) {
	&errmsg($@);
    }
}

# compare lists: &aeq($#a+1,@a,$#b+1,@b)
# from perlfunc(1)
sub aeq {   # compare two list values
    local(@a) = splice(@_,0,shift);
    local(@b) = splice(@_,0,shift);
    return 0 unless @a == @b;       # same len?
    while (@a) {
	return 0 if pop(@a) ne pop(@b);
    }
    return 1;
}

# exit. note that this does NOT check whether other than current buffers have
# been modified.
sub quit {
    my $prick;

    $diedie=1;

    if(!&aeq($#$cbuf+1,@$cbuf,$#{$$cinfo{origbuf}}+1,@{$$cinfo{origbuf}})) {
	$prick=&getline("Current buffer has been modified, quit anyway? ");
	if($prick!~/^\s*[Yy]/) {
	    &statmsg("Cancelled");
	    $diedie=0;
	}
    }
}
